unit ucosanubeseable;

{$mode delphi}

interface

uses
  Classes, SysUtils, ucosa, uConstantesSimSEE,
  unubeseable,u_uploadarchi,
  Dialogs;

type

  { TCosaNubeseable }

  TCosaNubeseable = class(TCosa)

    nid: integer;
    nid_version: integer;
    flg_nubeseable: boolean;

    constructor Create(capa: integer); override;

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText_(f: TArchiTexto); override;

    procedure WriteToText_BEGIN(f: TArchiTexto; const nombre: string); override;
    procedure WriteToText_END(f: TArchiTexto; const nombre: string); override;

    function Rec: TCosa_RecLnk; override;

    //inserta un nuevo elemento a la nube. si acualizar_version es true actualiza
    // la versión del nid_data ya existente, de lo contrario (false) inserta un nuevo
    //ni_data con versión=1 y devuelve en el nid_data el nid del elemento subido
    //La función devuelve el nid_version del elemento subido
    function NubesearCosa(acualizar_version: boolean; var nid_data: integer;
      idHilo: integer): integer;
  end;


    { TRec_nubeseable }

  TRec_nubeseable = class(TCosa)

    nid: integer;
    nid_version: integer;
    flg_nubeseable: boolean;

    constructor Create(capa, nid, nid_version: integer; flg_nubeseable:
      Boolean);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText_(f: TArchiTexto); override;

    procedure WriteToText_BEGIN(f: TArchiTexto; const nombre: string); override;
    procedure WriteToText_END(f: TArchiTexto; const nombre: string); override;

    function Rec: TCosa_RecLnk; override;

  end;

    { TArchiRef_Nubeseable }

  TArchiRef_Nubeseable = class (TCosa)

    nid: integer;
    nid_version: integer;
    flg_nubeseable: boolean;

    ArchiRef : TArchiRef;

    constructor Create(capa, nid, nid_version: integer; flg_nubeseable:
      Boolean; ArchiRef: TArchiRef);
        constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText_(f: TArchiTexto); override;

    procedure WriteToText_BEGIN(f: TArchiTexto; const nombre: string); override;
    procedure WriteToText_END(f: TArchiTexto; const nombre: string); override;

    function Rec: TCosa_RecLnk; override;

    //Inserta el achivo a la base
    function NubesearArchi(acualizar_version: boolean; var nid_data: integer;
      idHilo: integer; archi:string): integer;

    function testearYResolver: boolean;


  end;

procedure AlInicio;
procedure AlFinal;


implementation

{ TArchiRef_Nubeseable }

constructor TArchiRef_Nubeseable.Create(capa, nid, nid_version: integer;
  flg_nubeseable: Boolean; ArchiRef: TArchiRef);
begin
  inherited Create(capa);
  self.nid:=nid;
  self.nid_version:= nid_version;
  self.flg_nubeseable:= flg_nubeseable;
  self.ArchiRef := ArchiRef;
end;

constructor TArchiRef_Nubeseable.Create_ReadFromText(f: TArchiTexto);
var
  dir_cache: string;
begin
  inherited Create_ReadFromText(f);
  // aca fijarse si archivo esta en cache,
  //si no descargar alchivo .bin desde la base y guardarlo en cache

  dir_cache := uConstantesSimSEE.getDir_CacheArchi;

  if not DirectoryExists( dir_cache ) then
    CreateDir( dir_cache );

  if (not FileExists(dir_cache + DirectorySeparator+ self.ArchiRef.archi)) and
     (not FileExists(self.ArchiRef.archi)) then
    nubeAdmin.Download( 'nube_archi_bloques', self.nid, self.nid_version, dir_cache);

end;

procedure TArchiRef_Nubeseable.WriteToText_(f: TArchiTexto);
begin
  inherited WriteToText_(f);
end;

procedure TArchiRef_Nubeseable.WriteToText_BEGIN(f: TArchiTexto;
  const nombre: string);
begin
  inherited WriteToText_BEGIN(f, nombre);
end;

procedure TArchiRef_Nubeseable.WriteToText_END(f: TArchiTexto;
  const nombre: string);
begin
  inherited WriteToText_END(f, nombre);
end;

function TArchiRef_Nubeseable.Rec: TCosa_RecLnk;
var
  res: TCosa_RecLnk;
begin
  res:=inherited Rec;
  res.addCampoDef_archRef('nb_ArchiDatos',ArchiRef, 176 );
  res.addCampoDef('nb_nid', nid, 176); //identificador de la cosa en la BD
  res.addCampoDef('nb_nid_version', nid_version, 176); //identificador de la version de la cosa en la BD
  res.addCampoDef('nb_flg_nubeseable', flg_nubeseable, 176); //bandera que indica si la cosa es nubeseable
  Result := res;
end;

function TArchiRef_Nubeseable.NubesearArchi(acualizar_version: boolean;
  var nid_data: integer; idHilo: integer; archi:string): integer;
var
  errCod: Integer;
  memo: TStrings;
  memo_str: string;
  nid_version, k: integer;
begin

  if nubeAdmin.PASAPORTE<>'' then
  begin
    memo := self.asMemo(idHilo);
    for k := 0 to memo.Count - 1 do
        memo_str := memo_str + memo[k] + #13;
    if not acualizar_version then
      nid_version := 1; //como se incerta en nuevo elemento, la versión es 1

    errCod:= nubeAdmin.uploadArchi(nid_data, nid_version, errCod, archi,memo_str);
    if errCod < 0 then
      raise Exception.Create('Error' + unubeseable.ecToMsg(errCod));

    self.flg_nubeseable := True;
    self.nid := nid_data;
    self.nid_version := nid_version;


    //Guardar en la lista de archivos nubeseados, para que no lo empaquete junto con la sala
    ReferenciasAArchivosNubeseados.Add(Self.ArchiRef); //esto no funciona??????


    nubeAdmin.WriteToCacheArchi(Self.ArchiRef.archi,self.nid, self.nid_version);

    Result := nid_version;
  end
  else
  begin
     {$IFDEF simsee}
      ShowMessage('No tiene los permisos para realizar el insert');
     {$ENDIF}
      Result := -1;
  end;


end;

function TArchiRef_Nubeseable.testearYResolver: boolean;
var
  camino: string;

begin
  if FileExists(ArchiRef.archiRefStr) then
    Result := True
  else
  begin
    if self.flg_nubeseable then
       camino:= uConstantesSimSEE.getDir_CacheArchi
    else
       camino:= lista_caminos.Locate(self.ArchiRef.archiRefStr);
    if camino <> '' then
    begin
      ArchiRef.archiRefStr:= camino;
      result:= true;
    end
    else
      Result := false;
  end;
end;


{ TRec_nubeseable }

constructor TRec_nubeseable.Create(capa, nid, nid_version: integer;
  flg_nubeseable: Boolean);
begin
  inherited Create(capa);

   self.nid:=nid;
   self.nid_version:= nid_version;
   self.flg_nubeseable:= flg_nubeseable;

end;

constructor TRec_nubeseable.Create_ReadFromText(f: TArchiTexto);
var
  g: TArchiTexto;
  memo: TStringList;
  cosa: TCosa;
  consulta: string;
  errCod, k: integer;
  nid, nid_version,a: Integer;
  memo_str: String;
begin


  Create_ReadFromText_BEGIN(f);

  LeerCampos(f, 0, self.rec_lnk.Count - 1);

  nid := self.nid;
  nid_version := self.nid_version;

  //Se crea g a partir de la nube o a partir de cache.
  consulta := nubeAdmin.GetData(nid, nid_version, errCod);
  if errCod <> 0 then
    raise Exception.Create('Error' + unubeseable.ecToMsg(errCod));

  memo := TStringList.Create;
  memo.Delimiter:= #13;
  memo.Text:= consulta;

  cosa := self.Create_FromMemo(f.CatalogoReferencias, f.idHilo, memo);

  for k := 0 to memo.Count - 1 do
      memo_str := memo_str + memo[k] + #13;
  nubeAdmin.WriteToCache(nid, nid_version,memo_str); //cache local

  { g := TArchiTexto.CreateForRead(f.idHilo, f.CatalogoReferencias,
    getDir_Tmp + 'aux_clonar' + IntToStr(f.idHilo) + '.tmp',
    f.abortarEnError);

  LeerCampos(g, 0, self.rec_lnk.Count - 1);
   }

   self := TRec_nubeseable(cosa);
   self.flg_nubeseable:=true;
   self.nid := nid;
   self.nid_version := nid_version;



  //Create_ReadFromText_END(f);
end;

procedure TRec_nubeseable.WriteToText_(f: TArchiTexto);
begin
  inherited WriteToText_(f);
  //asi esta bien, que escriba nid, nidversion,flg y nombre del archivo.
end;

procedure TRec_nubeseable.WriteToText_BEGIN(f: TArchiTexto; const nombre: string);
begin
  inherited WriteToText_BEGIN(f, nombre);
end;

procedure TRec_nubeseable.WriteToText_END(f: TArchiTexto; const nombre: string);
begin
  inherited WriteToText_END(f, nombre);
end;

function TRec_nubeseable.Rec: TCosa_RecLnk;
var
  res: TCosa_RecLnk;
begin
  res := inherited Rec;
  res.addCampoDef('nb_nid', nid, 169); //identificador de la cosa en la BD
  res.addCampoDef('nb_nid_version', nid_version, 169); //identificador de la version de la cosa en la BD
  res.addCampoDef('nb_flg_nubeseable', flg_nubeseable, 169); //bandera que indica si la cosa es nubeseable
  Result := res;
end;


{ TCosaNubeseable }

constructor TCosaNubeseable.Create(capa: integer);
begin
  inherited Create;
  self.nid := 1;        // ojo aca NUBE, que le pases como parametro estos
   self.nid_version := 1;
   self.flg_nubeseable := False;
end;


constructor TCosaNubeseable.Create_ReadFromText(f: TArchiTexto);
var
  g: TArchiTexto;
  memo: TStringList;
  cosa: TCosa;
  consulta: string;
  errCod: integer;

begin
  Create_ReadFromText_BEGIN(f);

  LeerCampos(f, 0, 4);
  if self.flg_nubeseable then
  begin
    //se crea g a partir de la nube o a partir de cache, si la cosa es
    //Nubeseable (flg_Nubeseable=true) no se lee la cosa de f si no de g.
    consulta := nubeAdmin.GetData(self.nid, self.nid_version, errCod);
    if errCod <> 0 then
      raise Exception.Create('Error' + unubeseable.ecToMsg(errCod));

    memo := TStringList.Create;
    memo.Delimiter:= #13;
    memo.Text:= consulta;

    cosa := self.Create_FromMemo(f.CatalogoReferencias, f.idHilo, memo);
    cosa.asMemo(f.idHilo);

    g := TArchiTexto.CreateForRead(f.idHilo, f.CatalogoReferencias,
      getDir_Tmp + 'aux_clonar' + IntToStr(f.idHilo) + '.tmp',
      f.abortarEnError);

    LeerCampos(g, 7, self.rec_lnk.Count - 1);
    g.Free;
  end
  else
    LeerCampos(f, 5, self.rec_lnk.Count - 1);

  Create_ReadFromText_END(f);
end;


procedure TCosaNubeseable.WriteToText_(f: TArchiTexto);
var
  rec_nubeseable: TRec_nubeseable;
begin

  if not flg_nubeseable then
    EscirbirCampos(f, 0, rec_lnk.Count - 1)
  else
  begin
    // si es nubeseable, no escribo en f la cosanubeseable, si no que el recnubeseable
    rec_nubeseable := TRec_nubeseable.Create(self.capa, self.nid,
                   self.nid_version, self.flg_nubeseable);
    rec_nubeseable.WriteToText_(f);

  end
end;

procedure TCosaNubeseable.WriteToText_BEGIN(f: TArchiTexto; const nombre: string
  );
var
  rec_nubeseable: TRec_nubeseable;
begin
  if not flg_nubeseable then
     inherited WriteToText_BEGIN(f, nombre)
  else
  begin  // si es nubeseable, no escribo en f la cosanubeseable, si no que el recnubeseable
    rec_nubeseable := TRec_nubeseable.Create(self.capa, self.nid,
                   self.nid_version, self.flg_nubeseable);
    rec_nubeseable.WriteToText_BEGIN(f,nombre);

  end

end;

procedure TCosaNubeseable.WriteToText_END(f: TArchiTexto; const nombre: string);
var
  rec_nubeseable: TRec_nubeseable;
begin
  if not flg_nubeseable then
     inherited WriteToText_END(f, nombre)
  else
  begin  // si es nubeseable, no escribo en f la cosanubeseable, si no que el recnubeseable
    rec_nubeseable := TRec_nubeseable.Create(self.capa, self.nid,
                   self.nid_version, self.flg_nubeseable);
    rec_nubeseable.WriteToText_END(f,nombre);

  end
end;



function TCosaNubeseable.Rec: TCosa_RecLnk;
var
  res: TCosa_RecLnk;
begin
  res := inherited Rec;
  res.addCampoDef('nb_nid', nid, 169); //identificador de la cosa en la BD
  res.addCampoDef('nb_nid_version', nid_version, 169); //identificador de la version de la cosa en la BD
  res.addCampoDef('nb_flg_nubeseable', flg_nubeseable, 169); //bandera que indica si la cosa es nubeseable
  Result := res;
end;

function TCosaNubeseable.NubesearCosa(acualizar_version: boolean;
  var nid_data: integer; idHilo: integer): integer;
var
  memo: TStrings;
  memo_str: string;
  nid_version, k: integer;
begin

  if nubeAdmin.PASAPORTE<>'' then
  begin
    memo := self.asMemo(idHilo);
    for k := 0 to memo.Count - 1 do
    memo_str := memo_str + memo[k] + #13;
    if not acualizar_version then
    begin
      nid_data := nubeAdmin.WriteToNube(memo_str);
      nid_version := 1; //como se incerta en nuevo elemento, la versión es 1
    end
    else
    begin
      nid_version := nubeAdmin.WriteNewVersionToNube(nid_data, memo_str);
    end;

    self.flg_nubeseable := True;
    self.nid := nid_data;
    self.nid_version := nid_version;

    nubeAdmin.WriteToCache(nid, nid_version,memo_str);

    Result := nid_version;
  end
  else
  begin
     {$IFDEF simsee}
      ShowMessage('No tiene los permisos para realizar el insert');
     {$ENDIF}
      Result := -1;
  end;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(TRec_nubeseable.ClassName, TRec_nubeseable);
  registrarClaseDeCosa(TArchiRef_Nubeseable.ClassName, TArchiRef_Nubeseable);
end;

procedure AlFinal;
begin
end;


end.
