unit unubeseable;

{$mode delphi}

interface

(**
   Esta unidad administra el Almacen de Cosas tanto en el CacheLocal como en la Nube.

   Las Cosas que puede almacenar son una cadena de caracteres identificadas por
   un nid_cosa que identifica la Instancia de la cosa y el nid_version que identifica
   la versión de la instancia de la cosa.

**)

uses
  urosx,
  uDataSetGenerico, xmatdefs,
  usqlutils,
  Dialogs,
  Classes, SysUtils,uauxiliares
  , u_uploadarchi ,uConstantesSimSEE
  {$IFDEF simsee}
  ,StdCtrls
  {$ENDIF}
  ;

const
  // Códigos de error
  EC_fallo_conexion_url_nube = -1;
  EC_no_encontrada_en_la_nube = -2;
  EC_no_encontrada_en_cache = -3;
  EC_no_encontrada_en_ningun_lado = -4;
  EC_usuario_no_autenticado = -5;
  EC_usuario_o_clave_no_validos = -6;

  // Ruta de cache
  rutaCache = 'c:\simsee\basura\data_simsee';


type

  TRecPermisosNube = class

  end;

  { TNubeAdmin }

  TNubeAdmin = class
    id_hilo: integer;
    cache_dir: string;
    nube_url: string;
    PASAPORTE: string; // obtenido en el login
    nid_usr: integer;
    nid_grupos: TDAofNInt;

    constructor Create(id_hilo: integer; cache_dir, nube_url: string);

    // retorna TRUE si fue exitoso e inicializa el PASAPORTE
    // Llama a la función login del script url_nube y obtiene
    // un pasaporte válido.
    // Todas las funciones contra la nube deben pasar el parámetro
    // psp=PASAPORTE  que será chequedo por el script url_nube antes
    // de proceder a ejecutar.
    function Login(usuario, clave: string): boolean;

    // Llama a la función logut de url_nube y cancela la validéz del
    // PASAPORTE. Vacía la variable PASAPORTE indicando que no está logueado.
    function logout: boolean;

    function Permisos(nid_data: integer): TRecPermisosNube;


    // Inserta una nueva instancia de una cosa y retorna el nid_cosa asignado
    // si tiene éxito. Si se produce un error retorna 0 (cero).
    // La primer instancia de una cosa tiene nid_version = 1
    function WriteToNube(Data: string): integer;

    // baja de la nube el Data correspondiente a (nid, nid_version)
    // retorna vacío si falló y en codErr el error.
    function ReadFromNube(nid_data, nid_version: integer;
      var codErr: integer): string;

    function WriteToCache(nid_data, nid_version: integer; const Data: string): boolean;
    function ReadFromCache(nid_data, nid_version: integer;
      var codErr: integer): string;

    function uploadArchi(var nid, nid_version: Integer; var errCod: integer; archi,
      data: string): integer;

    function WriteToCacheArchi(const archi: string; nid, nid_version: integer
      ): boolean;

    procedure Download(tbl_archivo: string; nid, nid_version: integer;
      carpetaDestino: string='');

    // retornan la última versión ya sea de la nube o del cache y -1 si no encuentran
    function UltimaVersionFromNube(nid_data: integer): integer;
    function UltimaVersionFromCache(nid_data: integer): integer;

    // Escribe a la nube nueva versión del Data (nid_data) y si tienen éxito,
    // retorna un entero positivo con el número de versión asignado.
    // Si hay error retoran un negativo con el código de error.
    function WriteNewVersionToNube(nid_data: integer; const Data: string): integer;

    // retorna el path completo al archivo en el cache (cache_dir)/(nid_data)_v(nid_version).txt
    function ArchiCache(nid_data, nid_version: integer): string;

    // Elimina todo los objetos de cache
    function ClearCache: integer;

    // Elimina todos los objetos de cache asociados a un (nid_data)
    function ClearCacheOf(nid_data: integer): integer;

    // si lo tiene en el cache lo lee de cache sino lo busca en la nube
    // si falla retorna el string vacío y en errCod retorna el código de error
    // si tiene éxito la variable errCod se retora en CERO.
    function GetData(nid_data, nid_version: integer; var errCod: integer): string;

  end;

var
  nubeAdmin: TNubeAdmin;
  dbcon_rosx: TDBrosxCon;
  {$IFDEF simsee}
  memo_login: TMemo;
  {$ENDIF}

// retorna mensaje de error leible por humano.
function ecToMsg(errCod: integer): string;


implementation

function CrearSiNoExiste(carpeta: string): string;
begin
  if not DirectoryExists(carpeta) then
    try
      mkdir(carpeta)
    except
      on E: Exception do
        raise Exception.Create('Error al intentar crear la carpeta: ' + carpeta);
    end;
  Result := carpeta;
end;

// retorna mensaje de error leible por humano.
function ecToMsg(errCod: integer): string;
begin
  case errCod of
    EC_fallo_conexion_url_nube: Result := 'Falló conexión a la url_nube';
    EC_no_encontrada_en_la_nube: Result := 'Cosa no encontrada en la nube';
    EC_no_encontrada_en_cache: Result := 'Cosa no encontrada en el cache';
    EC_no_encontrada_en_ningun_lado: Result :=
        'Cosa no encontrada ni en la nube, ni en cache';
    EC_usuario_no_autenticado: Result := 'Usuario no autenticado';
    EC_usuario_o_clave_no_validos: Result := 'Usuario o clave no validos';
    else
      Result := 'ErrCod: ' + IntToStr(errCod) + ' DESCONOCIDO!.';
  end;
end;



{ TNubeAdmin }

constructor TNubeAdmin.Create(id_hilo: integer; cache_dir, nube_url: string);
begin
  self.id_hilo := id_hilo;
  self.cache_dir := cache_dir;
  self.nube_url := nube_url;
end;

function TNubeAdmin.Login(usuario, clave: string): boolean;
var
  sql: string;
  ds: TResultadoQuery;
  r: TDataRecord;
begin
  PASAPORTE := dbcon_rosx.sql_login(usuario,clave);
  if PASAPORTE<>'' then
  {$IFDEF simsee}
  ShowMessage('Login correcto');
  {$ENDIF}
end;

function TNubeAdmin.logout: boolean;
var
    res: Boolean;
begin
  res := dbcon_rosx.sql_logout(PASAPORTE);

end;

function TNubeAdmin.Permisos(nid_data: integer): TRecPermisosNube;
begin

end;

function TNubeAdmin.WriteToNube(Data: string): integer;
var
  sql: string;
  ds: TResultadoQuery;
  r: TDataRecord;
  nid_version, nid_data: Integer;
  th: TSQLTableHelper;

begin
  //nid_version, usuario, grupo

  nid_version := 1; //cuando se incerta por primera vez vale 1.
  nid_data := 0; // para que se inserte con el nid incremental que tenga la tabla

  Result := dbcon_rosx.sql_exec_pass(PASAPORTE,Data, nid_version,nid_data,0);


end;

function TNubeAdmin.ReadFromNube(nid_data, nid_version: integer;
  var codErr: integer): string;
var
  sql: string;
  ds: TResultadoQuery;
  r: TDataRecord;
begin


  ds := dbcon_rosx.sql_query_pass(PASAPORTE,nid_data,nid_version,'');

  if ds.EOF then
  begin
    codErr := EC_no_encontrada_en_la_nube;
    Result := '';
  end
  else
  begin
    r := ds.First;
    codErr := 0;  //0: ok
    Result := r.GetByNameAsString('data_memo');

  end;
end;

function TNubeAdmin.WriteToCache(nid_data, nid_version: integer;
  const Data: string): boolean;
var
  f: TextFile;
  archi: string;
begin
  archi := self.ArchiCache(nid_data, nid_version);
  if not fileexists(archi) then
  begin
    AssignFile(f, archi);
    Rewrite(f);
    WriteLn(f, Data);
    CloseFile(f);
  end;

end;

function TNubeAdmin.ReadFromCache(nid_data, nid_version: integer;
  var codErr: integer): string;
var
  f: TextFile;
  archi, linea, memo: string;
begin
  archi := self.ArchiCache(nid_data, nid_version);
  memo:= '';
  if fileexists(archi) then
  begin
    AssignFile(f, archi);
    reset(f);
    while not EOF(f) do
    begin
      readLn(f, linea);
      memo := memo + linea + #13;
    end;
    Close(f);
    Result := memo;
    codErr := 0;
  end
  else
  begin
    codErr := EC_no_encontrada_en_cache;
    Result := '';
  end;
end;


function TNubeAdmin.uploadArchi(var nid, nid_version: Integer; var errCod: integer;
  archi,data: string): integer;
var
  carpeta: string;
  zipname: string;
  res: integer;
  ext_str: string;
  lstArchis: TStringList;
begin
  Result:=-1212;
  if not FileExists(archi) then
    raise Exception.Create('No se encuentra el archivo de sala: ' + archi)
  else
  begin
    ext_str:= ExtractFileExt( archi );
    begin
      lstArchis:= TStringList.Create;
      lstArchis.add( archi + '_' + IntToStr(nid)+ '_'+ IntToStr(nid_version) );
      zipname:= ChangeFileExt( archi, '' ) + '_' + IntToStr(nid)+ '_'+ IntToStr(nid_version);
      if ext_str <> '.zip' then
      begin
        if FileExists( zipName+'.zip' ) then
          deletefile( zipName+'.zip' );
      end;
      if zipearListaDeArchivos( lstArchis, zipName ) <> 1 then
         raise Exception.Create('No fue posible comprimir el archivo.' );
      zipname:= zipname+'.zip';
    end;

    writeln('Subiendo archivo empaquetado.');
    res:= upload_archi(dbcon_rosx, 'nube_archi_bloques',nid, nid_version,zipname
          ,data,PASAPORTE );
    if res >= 0 then
      writeln('El archivo fue subido exitosamente. Nid asignado: '+IntToStr( res ) )
    else
      raise Exception.Create('Falló la subida del archivo. '+u_uploadarchi.UltimoError);

    Result:=res;

  end;
end;

function TNubeAdmin.WriteToCacheArchi(const archi: string; nid,nid_version: integer): boolean;

begin
 Download( 'nube_archi_bloques', nid, nid_version, uConstantesSimSEE.getDir_CacheArchi);
end;

procedure TNubeAdmin.Download(tbl_archivo: string; nid, nid_version: integer;
  carpetaDestino: string);
begin

  {$IFDEF simsee}
  if self.PASAPORTE ='' then
  begin
      ShowMessage('Debe realizar el Login');
      exit
  end;
  {$ENDIF}
  bajarCarpeta( dbcon_rosx,'nube_archi_bloques', nid,
                               nid_version, self.PASAPORTE, carpetaDestino);

end;

function TNubeAdmin.UltimaVersionFromNube(nid_data: integer): integer;

var
  sql: string;
  ds: TResultadoQuery;
  r: TDataRecord;

begin

  ds := dbcon_rosx.sql_query_pass(PASAPORTE,nid_data,1,'');

  if ds.EOF then
    Result := -1 //no hay última versión del archivo solicitado en la nube
  else
  begin
    r := ds.First;
    Result := r.GetByNameAsInt('nid_version');
  end;

end;

function TNubeAdmin.UltimaVersionFromCache(nid_data: integer): integer;
var
  memo: string;
  knid_version, codErr: integer;
begin
  memo := '';
  knid_version := 0;
  codErr := 0;

  while codErr = 0 do
  begin
    Inc(knid_version);
    memo := self.ReadFromCache(nid_data, knid_version, codErr);
  end;

  if knid_version = 1 then
    Result := -1
  else
    Result := knid_version - 1;

end;

function TNubeAdmin.WriteNewVersionToNube(nid_data: integer;
  const Data: string): integer;
var
  sql: string;
  ds: TResultadoQuery;
  r: TDataRecord;
  nid_version: integer;
  th: TSQLTableHelper;
begin

  //Ultima versión
  nid_version := self.UltimaVersionFromNube(nid_data);
  nid_version := nid_version + 1; //uno mas de la que había

  if dbcon_rosx.sql_exec_pass(PASAPORTE,Data, nid_version,nid_data,0)>0 then
    Result := nid_version
  else
    Result := EC_fallo_conexion_url_nube;

end;

function TNubeAdmin.ArchiCache(nid_data, nid_version: integer): string;
var
  carpeta_data, archivo: string;

begin

  carpeta_data := CrearSiNoExiste(rutaCache);
  archivo := carpeta_data + DirectorySeparator + IntToStr(nid_data) + '_v' +
    IntToStr(nid_version) + '.txt';

  Result := archivo;
end;



function TNubeAdmin.ClearCache: integer;
var
  SR: TSearchRec;
begin
  if FindFirst(rutaCache + '*.*', faArchive+faHidden+faReadOnly, SR)= 0 then
   repeat
     DeleteFile(rutaCache+'\'+SR.Name);     //NO FUNCIONA, ARREGLAR
   until FindNext(SR) <> 0;
 Result := 0; // que devuelvo??
end;

function TNubeAdmin.ClearCacheOf(nid_data: integer): integer;
var
  archi: string;
  ult_version, k : integer;
begin

  ult_version := self.UltimaVersionFromCache(nid_data);

  for k:=1 to ult_version do
  begin
    archi := self.ArchiCache(nid_data, k);
    if fileexists(archi) then
    begin
      DeleteFile(archi);
    end;
  end;

  Result := 0; // que devuelvo??

end;

function TNubeAdmin.GetData(nid_data, nid_version: integer;
  var errCod: integer): string;
var
  NoHayCache: boolean;
  memo: string;

begin
  memo := '';
  //fijarse primero en el cache si el actor se encuentra allí, sino
  //bajar de la nube

  memo := ReadFromCache(nid_data, nid_version, errCod);
  if errCod = EC_no_encontrada_en_cache then
    memo := ReadFromNube(nid_data, nid_version, errCod);

  if errCod = EC_no_encontrada_en_la_nube then
    errCod := EC_no_encontrada_en_ningun_lado;

  Result := memo;

end;



initialization
{$IFDEF NUBE}
  nubeAdmin := TNubeAdmin.Create(0, 'C:\SimSEE\basura', '/zeros/x/rosx_prueba.php');
  dbcon_rosx := TDBrosxCon.Create('', 'latorre.adme.com.uy', 80, '/zeros/x/rosx_prueba.php');
  dbcon_rosx.determinar_tipo_servidor;
{$ENDIF}

finalization
{$IFDEF NUBE}
  if dbcon_rosx <> nil then
    dbcon_rosx.Free;
{$ENDIF}
end.

